import os
import sys
import shutil
from time import time
from typing import Any, Optional, List
from functools import wraps
from subprocess import Popen, PIPE, STDOUT
from tempfile import mkdtemp
import multiprocessing as mp
import multiprocessing.pool
from autode.config import Config
from autode.log import logger
from autode.values import Allocation
from autode.exceptions import (NoAtomsInMolecule,
                               NoCalculationOutput,
                               NoConformers,
                               NoMolecularGraph,
                               MethodUnavailable)

try:
    mp.set_start_method("fork")
except RuntimeError:
    logger.warning('Multiprocessing context has already been defined')


if sys.version_info.minor > 7:                                    # Python >3.7
    from functools import cached_property # lgtm[py/unused-import]
    # LGTM alert is suppressed as this is imported in Python >3.7

if sys.version_info.minor <= 7:                                   # Python <3.7
    # Define a cached_property equivalent decorator

    class cached_property(object):
        # Based on https://github.com/pydanny/cached-property/

        def __init__(self, func):
            self.__doc__ = func.__doc__
            self.func = func

        def cached_property_wrapper(self, obj, _cls):
            if obj is None:
                return self

            value = obj.__dict__[self.func.__name__] = self.func(obj)
            return value

        __get__ = cached_property_wrapper


def check_sufficient_memory(func):
    """Decorator to check that enough memory is available for a calculation"""

    @wraps(func)
    def wrapped_function(*args, **kwargs):

        physical_mem = None
        required_mem = int(Config.n_cores) * Config.max_core

        try:
            physical_mem = Allocation(os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES'),
                                      units='bytes')
        except (ValueError, OSError):
            logger.warning('Cannot check physical memory')

        if physical_mem is not None and physical_mem < required_mem:
            raise RuntimeError('Cannot run function - insufficient memory. Had'
                               f' {physical_mem.to("GB")} GB but required '
                               f'{required_mem.to("GB")} GB')

        return func(*args, **kwargs)

    return wrapped_function


@check_sufficient_memory
def run_external(params, output_filename):
    """
    Standard method to run a EST calculation with subprocess writing the
    output to the calculation output filename

    ---------------------------------------------------------------------------
    Arguments:
        output_filename (str):

        params (list(str)): e.g. [/path/to/method, input-filename]
    """

    with open(output_filename, 'w') as output_file:
        # /path/to/method input_filename > output_filename
        process = Popen(params, stdout=output_file, stderr=PIPE)

        with process.stderr:
            for line in iter(process.stderr.readline, b''):
                logger.warning('STDERR: %r', line.decode())

        process.wait()

    return None


@check_sufficient_memory
def run_external_monitored(params, output_filename, break_word='MPI_ABORT',
                           break_words=None):
    """
    Run an external process monitoring the standard output and error for a
    word that will terminate the process

    ---------------------------------------------------------------------------
    Arguments:
        params (list(str)):

        output_filename (str):

    Keyword Arguments:
        break_word (str): String that if found will terminate the process

        break_words (list(str) | None): List of break_word-s
    """
    # Defining a set will override a single break word
    break_words = [break_word] if break_words is None else break_words

    def output_reader(process, out_file):
        for line in process.stdout:
            if any(word in line.decode('utf-8') for word in break_words):
                raise ChildProcessError

            print(line.decode('utf-8'), end='', file=out_file)

        return None

    with open(output_filename, 'w') as output_file:

        proc = Popen(params, stdout=PIPE, stderr=STDOUT)

        try:
            output_reader(proc, output_file)

        except ChildProcessError:
            logger.warning('External terminated')
            proc.terminate()
            return

    return None


def work_in(dir_ext):
    """Execute a function in a different directory"""

    def func_decorator(func):

        @wraps(func)
        def wrapped_function(*args, **kwargs):

            here = os.getcwd()
            dir_path = os.path.join(here, dir_ext)

            if not os.path.isdir(dir_path):
                logger.info(f'Creating directory to store files: {dir_path:}')
                os.mkdir(dir_path)

            os.chdir(dir_path)
            result = func(*args, **kwargs)
            os.chdir(here)

            if len(os.listdir(dir_path)) == 0:
                logger.warning(f'Worked in {dir_path} but made no files '
                               f'- deleting')
                os.rmdir(dir_path)

            return result

        return wrapped_function
    return func_decorator


def work_in_tmp_dir(filenames_to_copy: Optional[List[str]] = None,
                    kept_file_exts:    Optional[List[str]] = None,
                    use_ll_tmp:        bool = False):
    """Execute a function in a temporary directory.

    -----------------------------------------------------------------------
    Arguments:
        filenames_to_copy: Filenames to copy to the temp dir

        kept_file_exts: Filename extensions to copy back from the temp dir

        use_ll_tmp (bool): If true then use autode.config.Config.ll_tmp_dir
    """
    from autode.config import Config
    if filenames_to_copy is None:
        filenames_to_copy = []

    if kept_file_exts is None:
        kept_file_exts = []

    def func_decorator(func):

        @wraps(func)
        def wrapped_function(*args, **kwargs):
            here = os.getcwd()

            base_dir = Config.ll_tmp_dir if use_ll_tmp else None

            if base_dir is not None:
                assert os.path.exists(base_dir)

            tmpdir_path = mkdtemp(dir=base_dir)
            logger.info(f'Creating tmpdir to work in: {tmpdir_path}')

            if len(filenames_to_copy) > 0:
                logger.info(f'Copying {filenames_to_copy}')

            for filename in filenames_to_copy:
                if filename.endswith('_mol.in'):
                    # MOPAC needs the file to be called this
                    shutil.move(filename, os.path.join(tmpdir_path, 'mol.in'))
                else:
                    shutil.copy(filename, tmpdir_path)

            # Move directories and execute
            os.chdir(tmpdir_path)

            logger.info('Function   ...running')
            result = func(*args, **kwargs)
            logger.info('           ...done')

            for filename in os.listdir(tmpdir_path):

                if any([filename.endswith(ext) for ext in kept_file_exts]):
                    logger.info(f'Copying back {filename}')
                    shutil.copy(filename, here)

            os.chdir(here)

            logger.info('Removing temporary directory')
            shutil.rmtree(tmpdir_path)
            return result

        return wrapped_function
    return func_decorator


def log_time(prefix='Executed in: ', units='ms'):
    """A function requiring a number of atoms to run"""

    if units.lower() == 's' or units.lower() == 'seconds':
        s_to_units = 1.0

    elif units.lower() == 'ms' or units.lower() == 'milliseconds':
        s_to_units = 1000.0

    else:
        raise ValueError(f'Unsupported time unit: {units}')

    def func_decorator(func):

        @wraps(func)
        def wrapped_function(*args, **kwargs):
            start_time = time()

            result = func(*args, **kwargs)

            logger.info(f'{prefix} '
                        f'{(time() - start_time) * s_to_units:.2f} {units}')

            return result

        return wrapped_function
    return func_decorator


def requires_atoms(func):
    """A function requiring a number of atoms to run"""

    @wraps(func)
    def wrapped_function(*args, **kwargs):

        # Species must be the first argument
        assert hasattr(args[0], 'n_atoms')

        if args[0].n_atoms == 0:
            raise NoAtomsInMolecule

        return func(*args, **kwargs)

    return wrapped_function


def requires_graph(func):
    """A function requiring a number of atoms to run"""

    @wraps(func)
    def wrapped_function(*args, **kwargs):

        # Species must be the first argument
        assert hasattr(args[0], 'graph')

        if args[0].graph is None:
            raise NoMolecularGraph

        return func(*args, **kwargs)

    return wrapped_function


def requires_conformers(func):
    """A function requiring the species to have a list of conformers"""

    @wraps(func)
    def wrapped_function(*args, **kwargs):

        # Species must be the first argument
        assert hasattr(args[0], 'n_conformers')

        if args[0].n_conformers == 0:
            raise NoConformers

        return func(*args, **kwargs)

    return wrapped_function


def requires_hl_level_methods(func):
    """A function requiring both high and low-level methods to be available"""

    @wraps(func)
    def wrapped_function(*args, **kwargs):
        from autode.methods import get_lmethod, get_hmethod

        suffix = 'neither was available.'

        try:
            _ = get_lmethod()

            # Have a low-level method, so the high-level must not be available
            suffix = 'the high-level was not available.'
            _ = get_hmethod()

        except MethodUnavailable:
            raise MethodUnavailable(f'Function *{func.__name__}* requires both'
                                    f' a high and low-level method but '
                                    f'{suffix}')

        return func(*args, **kwargs)

    return wrapped_function


def requires_output(func):
    """A function requiring an output file and output file lines"""

    @wraps(func)
    def wrapped_function(*args, **kwargs):
        # Calculation must be the first argument
        assert hasattr(args[0], 'output')

        if args[0].output.file_lines is None:
            raise NoCalculationOutput

        return func(*args, **kwargs)

    return wrapped_function


def timeout(seconds, return_value=None):
    """
    Function dectorator that times-out after a number of seconds

    Arguments:
        seconds (float):

    Keyword Arguments:
        return_value (Any): Value returned if the function times out

    Returns:
        (Any): Result of the function | return_value
    """
    def handler(queue, func, args, kwargs):
        queue.put(func(*args, **kwargs))

    def decorator(func):

        def wraps(*args, **kwargs):
            q = multiprocessing.Queue()
            p = multiprocessing.Process(target=handler,
                                        args=(q, func, args, kwargs))

            if isinstance(mp.get_context(), mp.context.ForkContext):
                p.start()
            else:
                logger.error('Failed to wrap function')
                return func(*args, **kwargs)

            p.join(timeout=seconds)

            if p.is_alive():
                p.kill()
                p.join()
                return return_value

            else:
                return q.get()

        return wraps

    return decorator


def hashable(_method_name: str, _object: Any):
    """Multiprocessing requires hashable top-level functions to be executed,
    so convert a method into a top-level function"""
    return getattr(_object, _method_name)
